#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

#define osErrorFatal(userMsg) osErrorFatalImpl((userMsg), __FILE__, __func__, __LINE__)
#define osAssert(expr, userMsg) \
    do { \
        if (!(expr)) \
            osErrorFatal(userMsg); \
    } while(0)

static const char *os_Usage = "Usage: ./lock_file path/to/file start bytes sleepTimeInSec lockType (r or w)";
    
void osErrorFatalImpl(const char *userMsg, const char *fileName, 
                      const char *functionName, const int lineNum);

int main(int argc, char **argv) {
    osAssert(6 == argc, os_Usage);
    
    /*
     * parsiramo argumente i pretvaramo ih u odgovarajuce tipove
     */
    char *fpath = argv[1];
    int start = atoi(argv[2]);
    int length = atoi(argv[3]);
    int sleepTimeInSec = atoi(argv[4]);
    int lockType = argv[5][0] == 'r' ? F_RDLCK : F_WRLCK;
    
    /*
     * otvaramo fajl u kombinovanom modu, jer korisnik moze da postawi i r i w katanac.
     */
    int fd = open(fpath, O_RDWR);
    osAssert(-1 != fd, "Opening file failed");
	
	/* u strukturi flock moramo da navedemo svojstva katanca koji zelimo da postavimo */
    struct flock lock;
    lock.l_type = lockType;
    lock.l_whence = SEEK_SET;
    lock.l_start = start;
    lock.l_len = length;
    printf("Trying to lock...\n");
	/* pokusavamo da zakljucamo fajl na zeljeni nacin
	 * F_SETLK - ako ne moze da zakljuca fajl, odmah puca i vraca se nazad
	 * F_SETLKW - ako ne moze da zakljuca fajl, sacekace dok se resurs ne oslobodi
	 * F_GETLK - se koristi za ispitivanje da li moze da se postavi zeljeni katanac
	 *			 u ovom slucaju nije lose biti pohlepan i uvek ispitivati maksimalne mogucnosti, tj.
	 *           treba postaviti tip katanca na F_WRLCK. Ako fajl moze da se zakljuca u w modu, sigurno moze
	 *			 i u r modu. Vodite se parolom "ko moze vise moze i manje".
	 *
	 * Kada se vrsi zakljucavanje, uvek treba zakljucavati minimum potrebnih bajtova da bi se omogucio sto
	 * veci paralelizam. 
	 */
    osAssert(-1 != fcntl(fd, F_SETLK, &lock), "Locking file failed");
    printf("Lock acquired...\n");
    
    /*
     * Uspavljujemo proces cime simuliramo drzanje katanca
     */
    sleep(sleepTimeInSec);
	/* svaki postavljeni katanac je neophodno otkljucati u nekom trenutku 
	 * tip postavljamo na F_UNLCK
	 */
    lock.l_type = F_UNLCK;
	/* I otkljucavamo fajl */
    osAssert(-1 != fcntl(fd, F_SETLK, &lock), "Unlocking file failed");
    printf("Lock released...\n");
	/* zatvaramo fajl deskriptor
	 * i da zaboravite otkljucavanje, zatvaranjem deskriptora sa automatski oslobadaju svi katanci 
	 */
    close(fd);
	
    return 0;
}

void osErrorFatalImpl(const char *userMsg, const char *fileName, 
                      const char *functionName, const int lineNum) {
    perror(userMsg);
    fprintf(stderr, "File: '%s'\nFunction: '%s'\nLine: '%d'\n", fileName, functionName, lineNum);
    exit(EXIT_FAILURE);
}
