#define _XOPEN_SOURCE 700
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <sys/wait.h>

#define osErrorFatal(userMsg) osErrorFatalImpl((userMsg), __FILE__, __func__, __LINE__)
#define osAssert(expr, userMsg) \
    do { \
        if (!(expr)) \
            osErrorFatal(userMsg); \
    } while(0)

static const char *os_Usage = "Usage: ./stdio_syscalls_connection path/to/reg/file";
    
void osErrorFatalImpl(const char *userMsg, const char *fileName, 
                      const char *functionName, const int lineNum);

#define PIPE_RD (0)
#define PIPE_WR (1)
#define MAX_MSG_LEN (256)

int main(int argc, char **argv) {
    osAssert(2 == argc, os_Usage);
    
    printf("C STDIO to SYSCALLS demonstration:\n===================================\n");
    
    /*
     * otvaramo fajl pomocu fopen 
     */
    FILE *fileDem = fopen(argv[1], "r");
    osAssert(NULL != fileDem, "Opening file failed");
	/* iz postojeceg streama dohvatamo fajl deskriptor */
    int fileDemFd = fileno(fileDem);
    osAssert(-1 != fileDemFd, "Getting file descriptor from FILE* failed");
	/* upotrebom fstata nad izvucenim fajl deskriptorom citamo informacije o fajlu */
    struct stat finfo;
    osAssert(-1 != fstat(fileDemFd, &finfo), "Getting info through file descriptor failed");
    
    /*
     * Stampamo velicinu fajla
     */
    printf("File size ('%s'): %jd\n", argv[1], (intmax_t)finfo.st_size);
	/* Zatvaranjem fajl strima automatski se zatvara i fajl deskriptor kome je strim pridruzen, tj.
	 * dovoljno je pozvati fclose().
	 *
	 * BITNO:
	 * zatvaranje fajl deskriptora nad otvorenim strimom dovodi do nedefinisanog ponasanja programa
	 * i to ne treba nikada raditi
	 */
    fclose(fileDem);
    
    printf("\nSYSCALLS to C STDIO demnostration:\n===================================\n");
    
    /*
     * Kreiramo pajp
     */
    int pipeFds[2];
    osAssert(-1 != pipe(pipeFds), "Opening pipe failed");
	/* forkujemo procese */
    pid_t child = fork();
    osAssert(-1 != child, "Forking child failed");
    
    /*
     * Interprocesnu komunikaciju ostvarajuemo pomocu standardne biblioteke
	 * BITNO:
	 * read/write nema smisla koristiti kada se radi o formatiranom ulazu/izlazu jer nikada
	 * nemate garanciju gde ce se read/write prekinuti, tj. da li ce vam poziv puci na pola reci ili pola broja
	 * pa debagovanje postaje nocna mora zbog previse granicnih slucajeva.
	 * U tom slucaju mnogo je lakse i prirodnije koristiti standardnu biblioteku.
     */
    if (child) { // Parent
        const char *messagePrivateToParent = "Hello my child how are you doing?";
        const int magicNumber = 524287;
        
        /*
		 * Zatvaramo stranu koju ne koristimo
		 */
        close(pipeFds[PIPE_RD]);
		/* fajl deskriptoru pridruzujemo strim */
        FILE *pipeFileStream = fdopen(pipeFds[PIPE_WR], "w");
        osAssert(NULL != pipeFileStream, "Creating C stream from file descriptor failed (parent)");
		/* detetu saljemo formatirannu poruku preko standardne biblioteke */
        fprintf(pipeFileStream, "%s Magic: %d", messagePrivateToParent, magicNumber);
		/* zatvaranjems trima, automatski se zatvara i fajl deskriptor */
        fclose(pipeFileStream);
    } else { // Child
	
		/*
		 * Zatvaramo stranu koju ne koristimo
		 */
        close(pipeFds[PIPE_WR]);
        
        char buf[MAX_MSG_LEN];
		/* fajl deskriptoru pridruzujemo strim */
        FILE *pipeFileStream = fdopen(pipeFds[PIPE_RD], "r");
        osAssert(NULL != pipeFileStream, "Creating C stream from file descriptor failed (child)");
		/* citamo poruku od parenta pomocu standardne biblioteke */
        osAssert(NULL != fgets(buf, MAX_MSG_LEN, pipeFileStream), "Getting message failed");
		/* stampamo primljenu poruku na stdout */
        printf("Got: '%s'\n", buf);
        
        /* zatvaranjems trima, automatski se zatvara i fajl deskriptor */
        fclose(pipeFileStream);
		
		/* zavrsavamo child proces */
        exit(EXIT_SUCCESS);
    }
	
    /* cekamo da se child zavrsi i izbegavamo sejanje zombija po sistemu */
    osAssert(-1 != wait(NULL), "Waiting for child failed");
	
    return 0;
}

void osErrorFatalImpl(const char *userMsg, const char *fileName, 
                      const char *functionName, const int lineNum) {
    perror(userMsg);
    fprintf(stderr, "File: '%s'\nFunction: '%s'\nLine: '%d'\n", fileName, functionName, lineNum);
    exit(EXIT_FAILURE);
}
